/* hinting-viewer -- View font outline hinting
 * Copyright 2005, Red Hat
 * Author: Owen Taylor <otaylor@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <stdlib.h>
#include <string.h>

#include <X11/Xatom.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <gconf/gconf-client.h>

#include "render-area.h"

#define MAX_HISTORY 10

#define DEFAULT_TEXT "Hello"
#define DEFAULT_FONT "Sans 12"
#define DEFAULT_SCALE 20
#define DEFAULT_SHOW_PIXELS FALSE
#define DEFAULT_SHOW_OUTLINE TRUE
#define DEFAULT_OUTLINE_MODE FONT_OUTLINE_PLAIN
#define DEFAULT_HINT_MODE FONT_HINT_FULL

GtkWidget *font_button;

GSList *text_history = NULL;
GtkListStore *combo_store;

GConfClient *conf_client;

#define CONF_PREFIX "/apps/hinting-viewer"

#define FONT_KEY         CONF_PREFIX "/font"
#define TEXT_KEY         CONF_PREFIX "/text"
#define TEXT_HISTORY_KEY CONF_PREFIX "/text_history"
#define SCALE_KEY        CONF_PREFIX "/scale"
#define SHOW_PIXELS_KEY  CONF_PREFIX "/show_pixels"
#define SHOW_OUTLINE_KEY CONF_PREFIX "/show_outline"
#define HINT_MODE_KEY    CONF_PREFIX "/hint_mode"
#define OUTLINE_MODE_KEY CONF_PREFIX "/outline_mode"

static gboolean
get_bool_with_default (GConfClient *client,
		       const char  *key,
		       gboolean     dflt)
{
  GConfValue *value = gconf_client_get (client, key, NULL);
  gboolean retval = dflt;
  
  if (value)
    {
      if (value->type == GCONF_VALUE_BOOL)
	{
	  retval = gconf_value_get_bool (value);
	}

      gconf_value_free (value);
    }

  return retval;
}

static int
get_int_with_default (GConfClient *client,
		      const char  *key,
		      int          dflt)
{
  GConfValue *value = gconf_client_get (client, key, NULL);
  int retval = dflt;
  
  if (value)
    {
      if (value->type == GCONF_VALUE_INT)
	retval = gconf_value_get_int (value);

      gconf_value_free (value);
    }

  return retval;
}

static void
init_from_gconf (FontRenderArea *area)
{
  gboolean show_pixels;
  gboolean show_outline;
  FontHintMode hint_mode;
  FontOutlineMode outline_mode;
  PangoFontDescription *font_desc;
  char *str, *text, *font_name;
  int scale;

  /*
   * Text
   */
  text = gconf_client_get_string (conf_client, TEXT_KEY, NULL);
  if (!text)
    text = g_strdup (DEFAULT_TEXT);
  
  font_render_area_set_text (area, text);

  g_free (text);

  /*
   * Font
   */
  font_name = gconf_client_get_string (conf_client, FONT_KEY, NULL);
  if (!font_name)
    font_name = g_strdup (DEFAULT_FONT);

  font_desc = pango_font_description_from_string (font_name);
  font_render_area_set_font_description (area, font_desc);
  pango_font_description_free (font_desc);

  g_free (font_name);
  
  /*
   * Scale
   */
  scale = get_int_with_default (conf_client, SCALE_KEY, DEFAULT_SCALE);
  font_render_area_set_scale (area, scale);

  /*
   * Show Outline
   */
  show_outline = get_bool_with_default (conf_client, SHOW_OUTLINE_KEY, DEFAULT_SHOW_OUTLINE);
  font_render_area_set_show_outline (area, show_outline);
  
  /*
   * Show Pixels
   */
  show_pixels = get_bool_with_default (conf_client, SHOW_PIXELS_KEY, DEFAULT_SHOW_PIXELS);
  font_render_area_set_show_pixels (area, show_pixels);

  /*
   * Hint Mode
   */
  hint_mode = DEFAULT_HINT_MODE;
  str = gconf_client_get_string (conf_client, HINT_MODE_KEY, NULL);
  if (str)
    {
      if (strcmp (str, "none") == 0)
	hint_mode = FONT_HINT_NONE;
      else if (strcmp (str, "auto") == 0)
	hint_mode = FONT_HINT_AUTO;
      else if (strcmp (str, "full") == 0)
	hint_mode = FONT_HINT_FULL;
	
      g_free (str);
    }
  
  font_render_area_set_hint_mode (area, hint_mode);

  /*
   * Outline Mode
   */
  outline_mode = DEFAULT_OUTLINE_MODE;
  str = gconf_client_get_string (conf_client, OUTLINE_MODE_KEY, NULL);
  if (str)
    {
      if (strcmp (str, "plain") == 0)
	outline_mode = FONT_OUTLINE_PLAIN;
      else if (strcmp (str, "points") == 0)
	outline_mode = FONT_OUTLINE_POINTS;
      else if (strcmp (str, "controls") == 0)
	outline_mode = FONT_OUTLINE_CONTROLS;
	
      g_free (str);
    }
  
  font_render_area_set_outline_mode (area, outline_mode);
}

static void
update_text_history (void)
{
  GtkTreeModel *model = GTK_TREE_MODEL (combo_store);
  GtkTreeIter iter;
  gboolean have_row;
  GSList *new_history;
  GSList *new, *old;

  new_history = gconf_client_get_list (conf_client, TEXT_HISTORY_KEY,
				       GCONF_VALUE_STRING, NULL);

  have_row = gtk_tree_model_get_iter_first (model, &iter);
  old = text_history;
  new = new_history;
  
  while (old && new)
    {
      char *old_text = old->data;
      char *new_text = new->data;

      if (strcmp (old_text, new_text) == 0)
	{
	  old = old->next;
	  new = new->next;
	}
      else
	{
	  gtk_list_store_insert_before (combo_store, &iter, &iter);
	  gtk_list_store_set (combo_store, &iter, 0, new_text, -1);
	  new = new->next;
	}

      have_row = gtk_tree_model_iter_next (model, &iter);
    }

  while (old)
    {
      gtk_list_store_remove (combo_store, &iter);
      old = old->next;
    }

  while (new)
    {
      char *new_text = new->data;

      if (have_row)
	gtk_list_store_insert_before (combo_store, &iter, &iter);
      else
	gtk_list_store_append (combo_store, &iter);
      gtk_list_store_set (combo_store, &iter, 0, new_text, -1);
      new = new->next;
    }

  g_slist_foreach (text_history, (GFunc)g_free, NULL);
  g_slist_free (text_history);

  text_history = new_history;
}

static void
text_history_notify (GConfClient *client,
		     guint        cnxn_id,
		     GConfEntry  *entry,
		     gpointer     user_data)
{
  update_text_history ();
}

static void
update_text (GtkEntry       *entry,
	     FontRenderArea *area,
	     gboolean        add_to_history)
{
  const char *text = gtk_entry_get_text (entry);
  font_render_area_set_text (area, text);

  gconf_client_set_string (conf_client, TEXT_KEY, text, NULL);

  if (add_to_history)
    {
      GtkTreeModel *model = GTK_TREE_MODEL (combo_store);
      GtkTreeIter iter;
      gboolean have_row;
      GSList *tmp_history;

      for (have_row = gtk_tree_model_get_iter_first (model, &iter);
	   have_row;
	   have_row = gtk_tree_model_iter_next (model, &iter))
	{
	  const char *model_text;

	  gtk_tree_model_get (model, &iter,
			      0, &model_text,
			      -1);
	  if (strcmp (text, model_text) == 0)
	    return;
	}

      if (g_slist_length (text_history) == MAX_HISTORY)
	{
	  GSList *l = g_slist_last (text_history);
	  g_free (l->data);
	  text_history = g_slist_delete_link (text_history, l);
	}

      tmp_history = g_slist_prepend (text_history, (gpointer)text);
      
      gconf_client_set_list (conf_client, TEXT_HISTORY_KEY,
			     GCONF_VALUE_STRING, tmp_history, NULL);

      g_slist_delete_link (tmp_history, tmp_history);
    }
}

static void
on_combo_changed (GtkComboBox    *combo,
		  FontRenderArea *area)
{
  if (gtk_combo_box_get_active (combo) != -1)
    {
      GtkWidget *entry = gtk_bin_get_child (GTK_BIN (combo));
      
      update_text (GTK_ENTRY (entry), area, FALSE);
      gtk_editable_select_region (GTK_EDITABLE (entry), 0, -1);
    }
}

static void
on_entry_activate (GtkEntry *entry,
		   FontRenderArea *area)
{
  update_text (entry, area, TRUE);
}

static gboolean
on_entry_focus_out_event (GtkWidget      *widget,
			  GdkEventFocus  *event,
			  FontRenderArea *area)
{
  update_text (GTK_ENTRY (widget), area, TRUE);

  return FALSE;
}
  
static void
update_font (GtkFontButton  *font_button,
	     FontRenderArea *area)
{
  const char *font_name = gtk_font_button_get_font_name (font_button);
  PangoFontDescription *font_desc = pango_font_description_from_string (font_name);

  font_render_area_set_font_description (area, font_desc);

  gconf_client_set_string (conf_client, FONT_KEY, font_name, NULL);

  pango_font_description_free (font_desc);
}

static void
on_font_set (GtkFontButton *font_button,
	     FontRenderArea *area)
{
  update_font (font_button, area);
}


static void
do_save (GtkAction      *action,
	 FontRenderArea *area)
{
  
}

static void
do_save_as (GtkAction      *action,
	    FontRenderArea *area)
{
  
}

static void
do_select_font (GtkAction      *action,
		FontRenderArea *area)
{
  gtk_button_clicked (GTK_BUTTON (font_button));
}

static void
do_quit (GtkAction      *action,
	 FontRenderArea *area)
{
  gtk_main_quit ();
}

static void
do_zoom_in (GtkAction      *action,
	    FontRenderArea *area)
{
  int scale = area->scale;
  if (scale >= 20)
    scale += 5;
  else if (scale >= 10)
    scale += 2;
  else
    scale += 1;

  font_render_area_set_scale (area, scale);
  gconf_client_set_int (conf_client, SCALE_KEY, scale, NULL);
}

static void
do_zoom_out (GtkAction      *action,
	     FontRenderArea *area)
{
  int scale = area->scale;
  if (scale > 20)
    scale -= 5;
  else if (scale > 10)
    scale -= 2;
  else if (scale > 5)
    scale -= 1;
  
  font_render_area_set_scale (area, scale);
  gconf_client_set_int (conf_client, SCALE_KEY, scale, NULL);
}

static void
do_outline (GtkToggleAction *action,
	    FontRenderArea  *area)
{
  gboolean show_outline = gtk_toggle_action_get_active (action);
  
  font_render_area_set_show_outline (area, show_outline);
  gconf_client_set_bool (conf_client, SHOW_OUTLINE_KEY, show_outline, NULL);
}

static void
do_pixels (GtkToggleAction *action,
	    FontRenderArea  *area)
{
  gboolean show_pixels = gtk_toggle_action_get_active (action);
  
  font_render_area_set_show_pixels (area, show_pixels);
  gconf_client_set_bool (conf_client, SHOW_PIXELS_KEY, show_pixels, NULL);
}

static void
do_outline_mode (GtkRadioAction *action,
		 GtkRadioAction *current,
		 FontRenderArea *area)
{
  FontOutlineMode outline_mode = gtk_radio_action_get_current_value (action);
  const char *str = NULL;	/* Quiet GCC */
  
  font_render_area_set_outline_mode (area, outline_mode);

  switch (outline_mode)
    {
    case FONT_OUTLINE_PLAIN:
      str = "plain";
      break;
    case FONT_OUTLINE_POINTS:
      str = "plain";
      break;
    case FONT_OUTLINE_CONTROLS:
      str = "plain";
      break;
    }
  
  gconf_client_set_string (conf_client, OUTLINE_MODE_KEY, str, NULL);
}

static void
do_hint_mode (GtkRadioAction *action,
	      GtkRadioAction *current,
	      FontRenderArea *area)
{
  FontHintMode hint_mode = gtk_radio_action_get_current_value (action);
  const char *str = NULL;	/* Quiet GCC */
  
  font_render_area_set_hint_mode (area, hint_mode);

  switch (hint_mode)
    {
    case FONT_HINT_NONE:
      str = "none";
      break;
    case FONT_HINT_AUTO:
      str = "auto";
      break;
    case FONT_HINT_FULL:
      str = "full";
      break;
    }
  
  gconf_client_set_string (conf_client, HINT_MODE_KEY, str, NULL);
}

static void
get_workarea (GtkWidget    *widget,
	      GdkRectangle *workarea)
{
  GdkScreen *screen = gtk_widget_get_screen (widget);
  GdkDisplay *display = gdk_screen_get_display (screen);
  GdkWindow *root = gdk_screen_get_root_window (screen);
  Display *xdisplay = GDK_DISPLAY_XDISPLAY (display);
  Window xwindow = GDK_DRAWABLE_XID (root);
  gulong desktop = 0;

  Atom type;
  int format;
  gulong n_items, bytes_after;
  guchar *data;

  workarea->x = 0;
  workarea->y = 0;
  workarea->width = gdk_screen_get_width (screen);
  workarea->height = gdk_screen_get_height (screen);

  XGetWindowProperty (xdisplay, xwindow,
		      gdk_x11_get_xatom_by_name_for_display (display, "_NET_CURRENT_DESKTOP"),
		      0, G_MAXLONG, False,
		      XA_CARDINAL, &type, &format, &n_items, &bytes_after, &data);

  if (type == XA_CARDINAL)
    {
      if (format == 32 && n_items >= 1)
	{
	  gulong *long_data = (gulong *)data;
	  desktop = long_data[0];
	}

      XFree (data);
    }

  if ((4 * desktop) / 4 != desktop)
    return;

  XGetWindowProperty (xdisplay, xwindow,
		      gdk_x11_get_xatom_by_name_for_display (display, "_NET_WORKAREA"),
		      0, G_MAXLONG, False,
		      XA_CARDINAL, &type, &format, &n_items, &bytes_after, &data);

  if (type == XA_CARDINAL)
    {
      if (format == 32 && n_items >= 4 * desktop + 4)
	{
	  gulong *long_data = (gulong *)data;
	  GdkRectangle rect;
	  GdkRectangle intersect;

	  rect.x = MIN (long_data[4 * desktop], G_MAXINT);
	  rect.y = MIN (long_data[4 * desktop + 1], G_MAXINT);
	  rect.width = MIN (long_data[4 * desktop + 2], G_MAXINT);
	  rect.height = MIN (long_data[4 * desktop + 3], G_MAXINT);

	  gdk_rectangle_intersect (&rect, workarea, &intersect);
	  if (intersect.width > 0 || intersect.height > 0)
	    *workarea = intersect;
	}

      XFree (data);
    }
}

static void
snap_toplevel_size (GtkScrolledWindow *scrolled_window)
{
  GtkWidget *toplevel;
  GtkPolicyType hpolicy, vpolicy;
  GtkRequisition requisition;

  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (scrolled_window));
  if (toplevel && GTK_WIDGET_TOPLEVEL (toplevel))
    {
      GdkRectangle workarea;
      
      gtk_scrolled_window_get_policy (scrolled_window, &hpolicy, &vpolicy);

      gtk_scrolled_window_set_policy (scrolled_window, GTK_POLICY_NEVER, GTK_POLICY_NEVER);

      gtk_widget_size_request (toplevel, &requisition);

      get_workarea (GTK_WIDGET (scrolled_window), &workarea);
      workarea.width -= 10;
      workarea.height -= 27;

      if (requisition.width > workarea.width)
	{
	  GtkRequisition new;
	  
	  gtk_scrolled_window_set_policy (scrolled_window, GTK_POLICY_ALWAYS, GTK_POLICY_NEVER);
	  gtk_widget_size_request (toplevel, &new);
	  
	  requisition.width = MIN (requisition.width, workarea.width);
	  requisition.height = new.height;
	}

      if (requisition.height > workarea.height)
	{
	  gtk_window_maximize (GTK_WINDOW (toplevel));

	  gtk_scrolled_window_set_policy (scrolled_window, hpolicy, vpolicy);
	  scrolled_window->hscrollbar_visible = FALSE;
	  scrolled_window->vscrollbar_visible = FALSE;
	  
	  return;
	}

      gtk_scrolled_window_set_policy (scrolled_window, hpolicy, vpolicy);
      scrolled_window->hscrollbar_visible = FALSE;
      scrolled_window->vscrollbar_visible = FALSE;
      
      gtk_window_resize (GTK_WINDOW (toplevel), requisition.width, requisition.height);

      if (GTK_WIDGET_VISIBLE (toplevel))
	{
	  gint x, y;
	  
	  gtk_window_get_position (GTK_WINDOW (toplevel), &x, &y);

	  if (x + requisition.width > workarea.x + workarea.width ||
	      y + requisition.height > workarea.y + workarea.height)
	    {
	      gtk_window_move (GTK_WINDOW (toplevel),
			       CLAMP (x, workarea.x, workarea.x + workarea.width - requisition.width),
			       CLAMP (y, workarea.y, workarea.y + workarea.height - requisition.height));
	    }
	}
      
    }
}

static gboolean
snap_toplevel_size_idle (gpointer data)
{
  snap_toplevel_size (data);

  return FALSE;
}

static void
on_area_size_request (GtkWidget         *widget,
		      GtkRequisition    *requisition,
		      GtkScrolledWindow *scrolled_window)
{
  GtkRequisition *last_requisition = g_object_get_data (G_OBJECT (widget), "last-requisition");

  if (last_requisition &&
      (last_requisition->width != requisition->width ||
       last_requisition->height != requisition->height))
    g_idle_add_full (G_PRIORITY_HIGH, snap_toplevel_size_idle, scrolled_window, NULL);

  g_object_set_data_full (G_OBJECT (widget), "last-requisition",
			  g_memdup (requisition, sizeof (GtkRequisition)),
			  g_free);
}

const char *fixed_strings[][2] = {
  { "Arabic",    "السلام عليكم" },
  { "Bengali",   "ষাগতোম" },
  { "English",   "Hello" },
  { "Georgian",  "გამარჯობა" },
  { "Greek",     "Γειά σας" },
  { "Hebrew",    "שלום" },
  { "Hindi",     "नमस्ते, नमस्कार।" },
  { "Punjabi",   "ਪੁਂਜਾਬਿ)" },
  { "Russian",   "Здравствуйте!" },
  { "Tamil",     "வணக்கம்" },
  { "Thai",      "สวัสดีครับ, สวัสดีค่ะ" },
};

static GtkWidget *
create_text_combo (void)
{
  GtkWidget *combo;
  GtkTreeIter iter;
  GtkCellRenderer *renderer;
  int i;

  combo_store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);

  for (i = 0; i < G_N_ELEMENTS (fixed_strings); i++)
    {
      gtk_list_store_append (combo_store, &iter);
      gtk_list_store_set (combo_store, &iter,
			  0, fixed_strings[i][1],
			  1, fixed_strings[i][0],
			  -1);
    }
  
  combo = gtk_combo_box_entry_new_with_model (GTK_TREE_MODEL (combo_store), 0);
  gtk_cell_layout_clear (GTK_CELL_LAYOUT (combo));
  
  renderer = gtk_cell_renderer_text_new ();
  gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (combo), renderer, TRUE);
  gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (combo), renderer,
				  "text", 0,
				  NULL);
  
  renderer = gtk_cell_renderer_text_new ();

  g_object_set (renderer,
		"foreground", "#333333",
		"style", PANGO_STYLE_ITALIC,
		NULL);
  gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (combo), renderer, FALSE);
  gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (combo), renderer,
				  "markup", 1,
				  NULL);

  return combo;
}

static const GtkActionEntry actions[] = {
  { "FileMenu", NULL,           "_File" },
  { "Save",     GTK_STOCK_SAVE,    NULL, NULL, NULL, G_CALLBACK (do_save) },
  { "SaveAs",   GTK_STOCK_SAVE_AS, NULL, NULL, NULL, G_CALLBACK (do_save_as) },
  { "Quit",     GTK_STOCK_QUIT,    NULL, NULL, NULL, G_CALLBACK (do_quit) },
  { "ViewMenu", NULL,           "_View" },
  { "SelectFont", NULL,         "_Select Font", "<Control>f", NULL, G_CALLBACK (do_select_font) },
  { "EditMenu", NULL,           "_Edit" },
  { "ZoomIn",   GTK_STOCK_ZOOM_IN, NULL, "<Control>plus", NULL, G_CALLBACK (do_zoom_in) },
  { "ZoomOut",  GTK_STOCK_ZOOM_OUT, NULL, "<Control>minus", NULL, G_CALLBACK (do_zoom_out) },
};

static GtkToggleActionEntry options_actions[] = {
  { "Outline",  NULL,           "Out_line", "<Control>l", NULL, G_CALLBACK (do_outline), DEFAULT_SHOW_OUTLINE },
  { "Pixels",   NULL,           "Pi_xels",  "<Control>x", NULL, G_CALLBACK (do_pixels),  DEFAULT_SHOW_PIXELS },
};

static const GtkRadioActionEntry outline_mode_actions[] = {
  { "NoPoints",   NULL,  "_None",     "<Control>n", NULL, FONT_OUTLINE_PLAIN },
  { "Points",     NULL,  "_Points",   "<Control>p", NULL, FONT_OUTLINE_POINTS },
  { "Controls",   NULL,  "_Controls", "<Control>c", NULL, FONT_OUTLINE_CONTROLS }
};

static const GtkRadioActionEntry hint_mode_actions[] = {
  { "Unhinted",   NULL,  "_Unhinted",   "<Control>u", NULL, FONT_HINT_NONE },
  { "Autohinted", NULL,  "_Autohinted", "<Control>a", NULL, FONT_HINT_AUTO },
  { "Hinted",     NULL,  "_Hinted",     "<Control>h", NULL, FONT_HINT_FULL }
};

static const char *ui_string =
"<ui>"
"  <menubar name='MenuBar'>"
"    <menu action='FileMenu'>"
"      <menuitem action='Save'/>"
"      <menuitem action='SaveAs'/>"
"      <separator/>" 
"      <menuitem action='Quit'/>"
"    </menu>"
"    <menu action='EditMenu'>"
"      <menuitem action='SelectFont'/>"
"    </menu>"
"    <menu action='ViewMenu'>"
"      <menuitem action='ZoomIn'/>"
"      <menuitem action='ZoomOut'/>"
"      <separator/>" 
"      <menuitem action='Outline'/>"
"      <menuitem action='Pixels'/>"
"      <separator/>" 
"      <menuitem action='NoPoints'/>"
"      <menuitem action='Points'/>"
"      <menuitem action='Controls'/>"
"      <separator/>" 
"      <menuitem action='Unhinted'/>"
"      <menuitem action='Autohinted'/>"
"      <menuitem action='Hinted'/>"
"    </menu>"
"  </menubar>"
"  <toolbar name='SideBar'>"
"    <toolitem action='Outline'/>"
"    <toolitem action='Pixels'/>"
"    <separator/>" 
"    <toolitem action='NoPoints'/>"
"    <toolitem action='Points'/>"
"    <toolitem action='Controls'/>"
"    <separator/>" 
"    <toolitem action='Unhinted'/>"
"    <toolitem action='Autohinted'/>"
"    <toolitem action='Hinted'/>"
"  </toolbar>"
"  <toolbar name='TopBar'>"
"    <toolitem action='ZoomIn'/>"
"    <toolitem action='ZoomOut'/>"
"  </toolbar>"
"</ui>";

int main (int argc, char **argv)
{
  GError *error = NULL;
  GtkWidget *window;
  GtkActionGroup *action_group;
  GtkAction *action;
  GtkUIManager *ui;
  GtkWidget *sidebar;
  GtkWidget *topbar;
  GtkWidget *scrolled_window;
  FontRenderArea *area;
  GtkWidget *main_vbox;
  GtkWidget *main_hbox;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *combo;
  GtkWidget *entry;
  PangoFontDescription *font_desc;
  char *font_name;

  gtk_init (&argc, &argv);

  conf_client = gconf_client_get_default ();
  gconf_client_add_dir (conf_client, CONF_PREFIX,
			GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);

  area = FONT_RENDER_AREA (font_render_area_new ());
  init_from_gconf (area);

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  g_signal_connect (window, "destroy",
		    G_CALLBACK (gtk_main_quit), NULL);

  action_group = gtk_action_group_new ("Actions");
  gtk_action_group_add_actions (action_group,
				actions, G_N_ELEMENTS (actions),
				area);
  
  gtk_action_group_add_toggle_actions (action_group,
				       options_actions, G_N_ELEMENTS (options_actions),
				       area);
  
  action = gtk_action_group_get_action (action_group, "Outline");
  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), area->show_outline);
  action = gtk_action_group_get_action (action_group, "Pixels");
  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), area->show_pixels);

  gtk_action_group_add_radio_actions (action_group,
				      outline_mode_actions, G_N_ELEMENTS (outline_mode_actions),
				      area->outline_mode,
				      G_CALLBACK (do_outline_mode), area);
  gtk_action_group_add_radio_actions (action_group,
				      hint_mode_actions, G_N_ELEMENTS (hint_mode_actions),
				      area->hint_mode,
				      G_CALLBACK (do_hint_mode), area);
  
  ui = gtk_ui_manager_new ();
  gtk_ui_manager_insert_action_group (ui, action_group, 0);
  gtk_window_add_accel_group (GTK_WINDOW (window),
			      gtk_ui_manager_get_accel_group (ui));

  if (!gtk_ui_manager_add_ui_from_string (ui, ui_string, -1, &error))
    g_error ("building menus failed: %s", error->message);

  main_vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), main_vbox);

  gtk_box_pack_start (GTK_BOX (main_vbox),
		      gtk_ui_manager_get_widget (ui, "/MenuBar"),
		      FALSE, FALSE, 0);
  
  main_hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (main_vbox), main_hbox, TRUE, TRUE, 0);

  sidebar = gtk_ui_manager_get_widget (ui, "/SideBar");
  gtk_toolbar_set_orientation (GTK_TOOLBAR (sidebar), GTK_ORIENTATION_VERTICAL);
  gtk_toolbar_set_style (GTK_TOOLBAR (sidebar), GTK_TOOLBAR_TEXT);
  gtk_box_pack_start (GTK_BOX (main_hbox), sidebar, FALSE, FALSE, 0);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (main_hbox), vbox, TRUE, TRUE, 0);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

  topbar = gtk_ui_manager_get_widget (ui, "/TopBar");
  gtk_toolbar_set_style (GTK_TOOLBAR (topbar), GTK_TOOLBAR_ICONS);
  gtk_toolbar_set_show_arrow (GTK_TOOLBAR (topbar), FALSE);
  gtk_box_pack_start (GTK_BOX (hbox), topbar, FALSE, FALSE, 0);

  combo = create_text_combo ();
  gtk_box_pack_start (GTK_BOX (hbox), combo, TRUE, TRUE, 0);

  update_text_history ();
  gconf_client_notify_add (conf_client, TEXT_HISTORY_KEY,
			   text_history_notify, NULL, NULL, NULL);
  
  g_signal_connect (combo, "changed",
		    G_CALLBACK (on_combo_changed), area);
  
  entry = gtk_bin_get_child (GTK_BIN (combo));
  gtk_entry_set_text (GTK_ENTRY (entry), area->text);
  gtk_widget_grab_focus (entry);

  g_signal_connect (entry, "activate",
		    G_CALLBACK (on_entry_activate), area);
  g_signal_connect (entry, "focus-out-event",
		    G_CALLBACK (on_entry_focus_out_event), area);

  font_button = gtk_font_button_new ();
  gtk_font_button_set_show_style (GTK_FONT_BUTTON (font_button), TRUE);
  gtk_box_pack_start (GTK_BOX (hbox), font_button, FALSE, FALSE, 0);

  font_desc = font_render_area_get_font_description (area);
  font_name = pango_font_description_to_string (font_desc);
  gtk_font_button_set_font_name (GTK_FONT_BUTTON (font_button), font_name);
  g_free (font_name);

  g_signal_connect (font_button, "font-set",
		    G_CALLBACK (on_font_set), area);

  scrolled_window = gtk_scrolled_window_new (NULL, NULL);
  gtk_box_pack_start (GTK_BOX (vbox), scrolled_window, TRUE, TRUE, 0);

  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled_window), GTK_WIDGET (area));

  gtk_widget_show_all (gtk_bin_get_child (GTK_BIN (window)));

  g_signal_connect_after (area, "size-request",
			  G_CALLBACK (on_area_size_request), scrolled_window);
  
  snap_toplevel_size (GTK_SCROLLED_WINDOW (scrolled_window));

  gtk_widget_show (window);

  gtk_main ();

  return 0;
}
