/* hinting-viewer -- View font outline hinting
 * Copyright 2005, Red Hat
 * Author: Owen Taylor <otaylor@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <cairo.h>

#include <pango/pangoft2.h>

#include <gdk/gdkx.h>
#include <gtk/gtkdrawingarea.h>

#include "render-area.h"
#include "render.h"

#define SCALE 20
#define MARGIN 10

static void font_render_area_finalize (GObject *object);

static void     font_render_area_size_request (GtkWidget      *widget,
					       GtkRequisition *requsition);
static gboolean font_render_area_expose_event (GtkWidget      *widget,
					       GdkEventExpose *event);

struct _FontRenderAreaClass
{
  GtkDrawingAreaClass parent_class;
};

G_DEFINE_TYPE (FontRenderArea, font_render_area, GTK_TYPE_DRAWING_AREA)

static void
font_render_area_class_init (FontRenderAreaClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);

  object_class->finalize = font_render_area_finalize;

  widget_class->expose_event = font_render_area_expose_event;
  widget_class->size_request = font_render_area_size_request;
}

static void
substitute_func (FcPattern *pattern, gpointer data)
{
  FontRenderArea *area = data;

  FcPatternDel (pattern, FC_HINTING);
  FcPatternAddBool (pattern, FC_HINTING, area->hint_mode != FONT_HINT_NONE);
  FcPatternDel (pattern, FC_AUTOHINT);
  FcPatternAddBool (pattern, FC_AUTOHINT, area->hint_mode == FONT_HINT_AUTO);
}

static void
font_render_area_init (FontRenderArea *area)
{
  static const GdkColor white = { 0, 0xffff, 0xffff, 0xffff };
  
  area->text = NULL;
  area->scale = 20;
  area->show_pixels = TRUE;
  area->show_outline = FALSE;
  area->hint_mode = FONT_HINT_FULL;
  area->outline_mode = FONT_OUTLINE_PLAIN;
  
  area->font_map = pango_ft2_font_map_new ();
  pango_ft2_font_map_set_resolution (PANGO_FT2_FONT_MAP (area->font_map),
				     96., 96.);
  pango_ft2_font_map_set_default_substitute (PANGO_FT2_FONT_MAP (area->font_map),
					     substitute_func, area, NULL);
  
  area->context = pango_ft2_font_map_create_context (PANGO_FT2_FONT_MAP (area->font_map));

  gtk_widget_modify_bg (GTK_WIDGET (area), GTK_STATE_NORMAL, &white);
}

static void
area_ensure_layout (FontRenderArea *area)
{
  if (!area->layout)
    {
      area->layout = pango_layout_new (area->context);
      if (area->text)
	pango_layout_set_text (area->layout, area->text, -1);
    }
}

static void
area_clear_layout (FontRenderArea *area)
{
  if (area->layout)
    {
      g_object_unref (area->layout);
      area->layout = NULL;
    }
}

static void
font_render_area_finalize (GObject *object)
{
  FontRenderArea *area = FONT_RENDER_AREA (object);

  area_clear_layout (area);
  g_free (area->text);
  
  G_OBJECT_CLASS (font_render_area_parent_class)->finalize (object);
}

static void
font_render_area_size_request (GtkWidget      *widget,
			       GtkRequisition *requisition)
{
  FontRenderArea *area = FONT_RENDER_AREA (widget);
  int width, height;
  
  area_ensure_layout (area);
  
  pango_layout_get_pixel_size (area->layout, &width, &height);

  requisition->width = area->scale * width + 2 * MARGIN;
  requisition->height = area->scale * height + 2 * MARGIN;
}

#define FLOOR(x, mod) (x) - (((x) < 0) ? (((mod) + (x) % (mod)) % (mod)) : ((x) % (mod)))

static gboolean
font_render_area_expose_event (GtkWidget      *widget,
			       GdkEventExpose *event)
{
  FontRenderArea *area = FONT_RENDER_AREA (widget);
  
  cairo_t *context = gdk_cairo_create (event->window);
  FontRenderer *renderer;
  int pos;

  area_ensure_layout (area);
  
  /* Clear the background */
  cairo_set_source_rgb (context, 1.0, 1.0, 1.0);
  cairo_rectangle (context,
		   event->area.x, event->area.y, event->area.width, event->area.height);
  cairo_fill (context);

  renderer = font_renderer_new ();
  font_renderer_set_scale (renderer, area->scale);

  if (area->show_pixels)
    {
      cairo_move_to (context, MARGIN, MARGIN);
      font_renderer_draw_layout_pixels (renderer, context, area->layout);
    }
  
  cairo_set_source_rgb (context, 0.5, 0.5, 0.5);
  
  pos = MARGIN + FLOOR (event->area.x - MARGIN, area->scale);
  while (pos < event->area.x + event->area.width)
    {
      cairo_rectangle (context,
		       pos, event->area.y,
		       1,   event->area.height);
      cairo_fill (context);
      pos += area->scale;
    }

  pos = MARGIN + FLOOR (event->area.y - MARGIN, area->scale);
  while (pos < event->area.y + event->area.height)
    {
      cairo_rectangle (context,
		       event->area.x,      pos,
		       event->area.width,  1);
      cairo_fill (context);
      pos += area->scale;
    }

  if (area->show_outline)
    {
      cairo_move_to (context, MARGIN, MARGIN);
      cairo_set_source_rgba (context, 0.0, 0.5, 0.8, 0.5);
      font_renderer_fill_layout (renderer, context, area->layout);

      cairo_move_to (context, MARGIN, MARGIN);
      cairo_set_source_rgb (context, 0.0, 0.0, 0.0);
      font_renderer_stroke_layout (renderer, context, area->layout);
      
      if (area->outline_mode != FONT_OUTLINE_PLAIN)
	{
	  cairo_set_source_rgb (context, 1.0, 0.0, 0.0);
	  cairo_set_line_width (context, 1.0);

	  cairo_move_to (context, MARGIN, MARGIN);
	  font_renderer_draw_layout_points (renderer, context, area->layout,
					    area->outline_mode == FONT_OUTLINE_CONTROLS);
	}
    }
  
  font_renderer_free (renderer);
  
  cairo_destroy (context);

  return FALSE;
  
}

GtkWidget*
font_render_area_new (void)
{
  return g_object_new (FONT_TYPE_RENDER_AREA, NULL);
}

void
font_render_area_set_text (FontRenderArea *area,
			   const char     *text)
{
  char *tmp;
  
  g_return_if_fail (FONT_IS_RENDER_AREA (area));

  tmp = g_strdup (text);
  g_free (area->text);
  area->text = tmp;

  area_clear_layout (area);
  gtk_widget_queue_resize (GTK_WIDGET (area));
}

void
font_render_area_set_font_description (FontRenderArea *area,
				       PangoFontDescription *font_desc)
{
  g_return_if_fail (FONT_IS_RENDER_AREA (area));

  pango_context_set_font_description (area->context, font_desc);
  
  area_clear_layout (area);
  gtk_widget_queue_resize (GTK_WIDGET (area));
}

PangoFontDescription *
font_render_area_get_font_description (FontRenderArea *area)
{
  g_return_val_if_fail (FONT_IS_RENDER_AREA (area), NULL);

  return pango_context_get_font_description (area->context);
}

void
font_render_area_set_scale (FontRenderArea *area,
			    int             scale)
{
  g_return_if_fail (FONT_IS_RENDER_AREA (area));
  g_return_if_fail (scale > 0);

  if (scale != area->scale)
    {
      area->scale = scale;
      gtk_widget_queue_resize (GTK_WIDGET (area));
    }
}

void
font_render_area_set_hint_mode (FontRenderArea *area,
				FontHintMode    hint_mode)
{
  g_return_if_fail (FONT_IS_RENDER_AREA (area));

  if (hint_mode != area->hint_mode)
    {
      area->hint_mode = hint_mode;

      pango_ft2_font_map_substitute_changed (PANGO_FT2_FONT_MAP (area->font_map));
      
      area_clear_layout (area);
      gtk_widget_queue_resize (GTK_WIDGET (area));
    }
}

void
font_render_area_set_outline_mode (FontRenderArea *area,
				   FontOutlineMode    outline_mode)
{
  g_return_if_fail (FONT_IS_RENDER_AREA (area));

  if (outline_mode != area->outline_mode)
    {
      area->outline_mode = outline_mode;
      gtk_widget_queue_draw (GTK_WIDGET (area));
    }
}

void
font_render_area_set_show_outline (FontRenderArea *area,
				   gboolean        show_outline)
{
  g_return_if_fail (FONT_IS_RENDER_AREA (area));

  show_outline = show_outline != FALSE;

  if (show_outline != area->show_outline)
    {
      area->show_outline = show_outline;
      gtk_widget_queue_draw (GTK_WIDGET (area));
    }
}

void
font_render_area_set_show_pixels (FontRenderArea *area,
				  gboolean        show_pixels)
{
  g_return_if_fail (FONT_IS_RENDER_AREA (area));

  show_pixels = show_pixels != FALSE;

  if (show_pixels != area->show_pixels)
    {
      area->show_pixels = show_pixels;
      gtk_widget_queue_draw (GTK_WIDGET (area));
    }
}
